from arrow_mssql.input.datatypes import (
    get_type_arrow, 
    map_typs, 
    get_type_arrow_pyodbc, 
    map_typs_pyodbc
)
import pyarrow as pa
import textwrap
from arrow_mssql.utils import rename_col


def drop_table(tbl: str) -> str:
    return f'DROP TABLE IF EXISTS {tbl}'


def create_table(
    tbl: str, 
    schema_arrow: pa.Schema
) -> str:
    
    list_types = []

    for field in schema_arrow:
        col = rename_col(field.name)
        tp = field.type
        null_type = '' if field.nullable == True else 'not null'

        if convert := get_type_arrow(tp):
           ty_out = convert
        else:
           ty_out = map_typs[tp]

        list_types.append(f'{col} {ty_out} {null_type}')
    
    stmt = f'CREATE TABLE {tbl} (\n'
    stmt += textwrap.indent(',\n'.join(map(str.strip,list_types)), '   ')
    stmt += '\n)'

    return stmt


def insert_setinputsizes(
    schema_arrow: pa.Schema
) -> list[tuple]:
    
    list_types = []

    for field in schema_arrow:
        tp = field.type

        if convert := get_type_arrow_pyodbc(tp):
           ty_out = convert
        else:
           ty_out = map_typs_pyodbc[tp]

        list_types.append(ty_out)

    return list_types


def insert_table(
    tbl: str, 
    schema_arrow: pa.Schema
) -> str:

    cols = [*map(lambda f: f.name, schema_arrow)]

    values = f"INSERT INTO {tbl} ({', '.join(cols)})"
    values += ' VALUES ('
    values += '?, ' * len(cols)
    values = values[:-2] + ')'

    return values


import pyarrow as pa
from uuid import UUID


class UuidScalarType(pa.ExtensionScalar):
    def as_py(self):
        return None if self.value is None else UUID(bytes=self.value.as_py())


class UuidType(pa.ExtensionType):
    def __init__(self):
        super().__init__(pa.binary(16), 'pacote.UuidType')

    def __arrow_ext_scalar_class__(self):
        return UuidScalarType

    def __arrow_ext_serialize__(self):
        return b''

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return cls()


def decimal_arrow(
    precision, 
    scale
) -> pa.DataType:
    
    if precision > 38:
        dtype = pa.decimal256(precision, scale)
    else:
        dtype = pa.decimal128(precision, scale)
    return dtype


def float_arrow(
    precision
) -> pa.DataType:
    
    if precision <= 24:
        return pa.float32()
    else:
        return pa.float64()


def time_arrow(
    precision
):
    if precision is None or precision == 0:
        return pa.time32('s')
    elif 1 <= precision <= 3:
        return pa.time32('ms')
    elif 4 <= precision <= 6:
        return pa.time64('us')
    elif 7 <= precision <= 9:
        return pa.time64('ns')


def timestamp_arrow(
    precision
) -> pa.DataType:
    
    if precision is None or precision == 0:
        return pa.timestamp('s')
    elif 1 <= precision <= 3:
        return pa.timestamp('ms')
    elif 4 <= precision <= 6:
        return pa.timestamp('us')
    elif 7 <= precision <= 9:
        return pa.timestamp('ns')


# NOTE: Ainda em testes @EXPERIMENTAL
UUID_TYPE_ARROW = UuidType()
pa.register_extension_type(UUID_TYPE_ARROW)

map_typs = {
    'int': pa.int32(),
    'bit': pa.bool_(),
    'char': pa.string(),
    'nchar': pa.string(),
    'text': pa.string(),
    'ntext': pa.string(),
    'binary': pa.binary(),
    'image': pa.binary(),
    'varbinary': pa.binary(),
    'timestamp': pa.binary(),
    'datetime': timestamp_arrow,
    'datetime2': timestamp_arrow,
    'smalldatetime': timestamp_arrow,
    'datetimeoffset': timestamp_arrow,
    'numeric': decimal_arrow,
    'decimal': decimal_arrow,
    'smallint': pa.int16(),
    'varchar': pa.string(),
    'nvarchar': pa.string(),
    'date': pa.date64(),
    'float': float_arrow,
    'real': pa.float32(),
    'bigint': pa.int64(),
    'money': pa.int64(),
    'smallmoney': pa.int32(),
    'tinyint': pa.int8(),
    'time': time_arrow,
    'uniqueidentifier': pa.string()
}


import pyodbc as odbc
from typing import Any
from contextlib import closing
import contextlib
import datetime
import struct


def datetimeoffset_to_datetime(value):
    (
        year, 
        month, 
        day, 
        hour, 
        minute, 
        second, 
        frac, 
        tz_hour, 
        tz_minutes
    ) = struct.unpack("<6hI2h", value)

    return datetime.datetime(
        year,
        month,
        day,
        hour,
        minute,
        second,
        frac // 1000,
        datetime.timezone(datetime.timedelta(hours=tz_hour, minutes=tz_minutes)),
    )


def do_connect(
    *args,
    **kwargs: Any
) -> None:
    
    con = odbc.connect(
        *args,
        **kwargs,
    )

    con.add_output_converter(
        -155, 
        datetimeoffset_to_datetime
    )
    
    if 'autocommit' not in kwargs:
        con.execute(
            'set transaction '
            'isolation level '
            'read uncommitted;'
        )

    with closing(con.cursor()) as cur:
        cur.execute("SET DATEFIRST 1")

    return con


@contextlib.contextmanager
def raw_sql(*args,**kwargs):

    con = do_connect(*args, **kwargs)
    cursor = con.cursor()

    try:
        yield cursor
    except Exception:
        raise
    finally:
        cursor.close()


import sqlglot as sg
import sqlglot.expressions  as sge
import unicodedata


def is_query(texto: str) -> bool:
    if sg.parse_one(texto, dialect='tsql').find(sge.Select):
        return True
    return False


def remove_accet(txt: str) -> str:
    t_01 = unicodedata.normalize('NFD', txt)
    t_02 = ''.join(c for c in t_01 if not unicodedata.combining(c))
    return unicodedata.normalize('NFC', t_02)


def remove_especial(text: str) -> str:
    import re
    return re.sub(r'[^a-zA-Z0-9 _]', '', text)


def rename_col(col: str) -> str:
    return '_'.join(
        remove_especial(
            remove_accet(c.strip())
        )
        .casefold() 
        for c in col.split()
    )


import pyarrow.parquet as pq
from arrow_mssql.connector import raw_sql
import contextlib
from pathlib import Path
import arrow_mssql.input.schema as db
from typing import (
    Generator,
    Any
)


@contextlib.contextmanager
def write_parquet(
    driver: str,
    name: str,
    *,
    path: str | Path,
    override: bool = True,
    limit: int = 10_000
) -> Generator[Any, Any, None]:
    

    with contextlib.suppress(AttributeError):
        if isinstance(path, str):
            path = Path(path)

        tbl = pq.ParquetFile(path)

        droptable = db.drop_table(name)
        create = db.create_table(name, tbl.schema_arrow)
        insert = db.insert_table(name, tbl.schema_arrow)
        insert_puts = db.insert_setinputsizes(tbl.schema_arrow)

    
    with raw_sql(driver, autocommit=False) as cursor:

        if override:
            cursor.execute(droptable)
            cursor.execute(create)
            cursor.commit()

        cursor.fast_executemany = True
        cursor.setinputsizes(insert_puts)

        for rows in tbl.iter_batches(limit):
            lotes = [tuple(d.values()) for d in rows.to_pylist()]
            cursor.executemany(insert, lotes)

        cursor.commit()
        
        yield cursor


import pyarrow as pa
from .connector import raw_sql
from textwrap import dedent
from typing import (
    Iterable, 
)
from .output.schemas import get_schema
from .utils import is_query
import sqlglot as sg


def cursor_arrow(
    driver: str,
    name: str,
    database: str,
    schema: str = 'dbo',
    limit: int = 1_000_000
) -> Iterable[list]:
    
    with raw_sql(driver) as cursor:
        stmt = (
            sg.select('*')
            .from_(
                sg.table(
                    f'{name}',
                    db=schema,
                    catalog=database
                )
            )
            .sql('tsql')
        )

        if is_query(name):
            stmt = dedent(name)

        cursor.execute(stmt)
        while batch := cursor.fetchmany(limit):
            yield batch


def to_arrow_lotes(
    driver: str,
    name: str,
    database: str,
    schema: str = 'dbo',
    limit: int = 1_000_000
) -> pa.ipc.RecordBatchReader:
    
    schema_arrow = get_schema(
        driver, 
        name, 
        database, 
        schema
    )
    array_type = schema_arrow('st')
    
    arrays = (
        pa.array(map(tuple, lote), type=array_type)
        for lote in cursor_arrow(
            driver, 
            name, 
            database, 
            schema,
            limit   
        )
    )

    lotes = map(
        pa.RecordBatch.from_struct_array, 
        arrays
    )

    return pa.ipc.RecordBatchReader.from_batches(
        schema_arrow('sh'), 
        lotes
    )


def to_parquet(
    driver: str,
    name: str,
    *,
    path: str,
    database: str,
    schema: str = 'dbo',
    row_group_size: int = 1_000_000,
    limit: int = 1_000_000
) -> None:
    """
    ### Exporta tabela ou consulta para .parquet
    
    ----
    driver: String de conexao `pyodbc`
    name: Tabela ou consulta
    path: Caminho do arquivo
    database: Banco de dados da tabela
    schema: schema da tabela
    row_group_size: Grupo do arquivo .parquet
    limit: porcao de dados
    """
    
    import pyarrow.parquet as pq
    
    with to_arrow_lotes(
        driver, 
        name, 
        database, 
        schema, 
        limit
    ) as lotes:
        
        with pq.ParquetWriter(path, lotes.schema) as writer:
            for lote in lotes:
                writer.write_batch(
                    lote, 
                    row_group_size=row_group_size
                )


def to_csv(
    driver: str,
    name: str,
    *,
    path: str,
    database: str,
    schema: str = 'dbo',
    delimiter: str = ';',
    limit: int = 1_000_000
) -> None:
    """
    ### Exporta tabela ou consulta para .csv
    
    ----
    driver: String de conexao `pyodbc`
    name: Tabela ou consulta
    path: Caminho do arquivo
    database: Banco de dados da tabela
    schema: schema da tabela
    delimiter: separador das colunas
    limit: porcao de dados
    """
    
    from pyarrow import csv
    
    with to_arrow_lotes(
        driver, 
        name, 
        database, 
        schema, 
        limit
    ) as lotes:
        
        write_options = csv.WriteOptions(
            include_header=True,
            delimiter=delimiter, 
            quoting_style='all_valid'
        )
        
        with csv.CSVWriter(
            path, 
            lotes.schema, 
            write_options=write_options
        ) as writer:
            for lote in lotes:
                writer.write_batch(lote)


import pyarrow as pa
from ..connector import raw_sql
from textwrap import dedent
from typing import (
    Any, 
    Literal, 
    Callable
)
from .datatypes import map_typs
from operator import itemgetter
import sqlglot as sg
import sqlglot.expressions as sge
from ..utils import (
    is_query,
    rename_col
)


def schema_query_or_table(
    driver: str,
    name: str,
    database: str,
    schema: str
) -> list[tuple]:
    
    query = is_query(name)

    if query:
       name = sge.convert(str(name)).sql('tsql')
       stmt = dedent(
           f'''
           EXEC sp_describe_first_result_set @tsql = N{name}
           '''
       )

    else:
        where = [
            sg.column("table_name").eq(sge.convert(name)),
            sg.column("table_schema").eq(sge.convert(schema))
        ]

        stmt = (
            sg.select(
                "column_name",
                "data_type",
                "is_nullable",
                "numeric_precision",
                "numeric_scale",
                "datetime_precision"
            )
            .from_(
                sg.table(
                    'columns',
                    db='information_schema',
                    catalog=database
                )
            )
            .where(*where)
            .order_by("ordinal_position")
            .sql('tsql')
        )

    with raw_sql(driver) as cur:
        rst = cur.execute(stmt).fetchall()
    
    if query:

        def alter_tuple(row):
            row = list(row)
            tipo = row[1]

            if '(' in tipo:
                row[1] = tipo[:tipo.find('(')]

            row[2] = 'YES' if row[2] else 'NO'
            return tuple(row)

        filtro = itemgetter(2, 5, 3, 7, 8, 8)
        rst = [
               *map(
                   alter_tuple, 
                   map(
                       filtro, 
                       rst
                )
            )
        ]
    
    return rst
    

def get_schema(
    driver: str,
    name: str,
    database: str,
    schema: str = 'dbo'
) -> Callable:

    
    rst = schema_query_or_table(
        driver,
        name,
        database,
        schema
    )
    
    def tipo(
        retorno: Literal['sh', 'st', 'lt']
    ) -> Any:
        
        if retorno == 'lt':
            return rst
        
        schema = []

        for (
            cols,
            types,
            is_null,
            precision,
            scale,
            dt_precision
        ) in rst:

            dtype = map_typs[types]     
            if types in ('decimal', 'numeric'):
                dtype = dtype(precision, scale)
            
            if types == 'float':
                dtype = dtype(precision)
            
            if types in (
                'datetime',
                'datetime2', 
                'smalldatetime',
                'datetimeoffset',
                'time'
            ):
                dtype = dtype(dt_precision)

            field = pa.field(
                rename_col(cols), 
                dtype, 
                nullable= is_null == 'YES'
            )
            schema.append(field)
            
        if retorno == 'st':
            return pa.struct(schema)        
        return pa.schema(schema)
    
    return tipo


import pyarrow as pa
import pyodbc


def decimal_mssql(tp: pa.DataType) -> str:
    
    precision = tp.precision
    scale = tp.scale

    if precision <= 28 and scale is None:
        return f'numeric({precision})'
    elif precision <=28 and scale >= 0:
        return f'decimal({precision}, {scale})'
    elif precision > 38:
        return f'decimal({precision}, {scale})'
    

def decimal_mssql_pyodbc(tp: pa.DataType) -> tuple:
    
    precision = tp.precision
    scale = tp.scale

    if precision <= 28 and scale is None:
        return (pyodbc.SQL_NUMERIC, precision, 0)
    elif precision <=28 and scale >= 0:
        return (pyodbc.SQL_DECIMAL, precision, scale)
    elif precision > 38:
        return (pyodbc.SQL_DECIMAL, precision, scale)


def timestamp_mssql(tp: pa.DataType) -> str:
    unit = tp.unit

    if unit == 's':
        return 'smalldatetime'
    elif unit == 'ms':
        return 'datetime'
    elif unit == 'us':
        # 'datetimeoffset'
        return 'datetime2'
    elif unit == 'ns':
        return 'datetime2'
    

def timestamp_mssql_pyodbc(tp: pa.DataType) -> tuple:
    unit = tp.unit

    if unit == 's':
        return (pyodbc.SQL_TYPE_TIMESTAMP, 0, 0)
    elif unit == 'ms':
        return (pyodbc.SQL_TYPE_TIMESTAMP, 3, 0)
    elif unit == 'us':
        return (pyodbc.SQL_TYPE_TIMESTAMP, 6, 0)
    elif unit == 'ns':
        return (pyodbc.SQL_TYPE_TIMESTAMP, 7, 0)


def times_mssql(tp: pa.DataType) -> str:
    unit = tp.unit

    if unit == 's':
        return 'time'
    elif unit == 'ms':
        return 'time(3)'
    elif unit == 'us':
        return 'time(6)'
    elif unit == 'ns':
        return 'time(7)'
    

def times_mssql_pyodbc(tp: pa.DataType) -> str:
    unit = tp.unit

    if unit == 's':
        return (pyodbc.SQL_TYPE_TIME, 0, 0)
    elif unit == 'ms':
        return (pyodbc.SQL_TYPE_TIME, 3, 0)
    elif unit == 'us':
        return (pyodbc.SQL_TYPE_TIME, 6, 0)
    elif unit == 'ns':
        return (pyodbc.SQL_TYPE_TIME, 7, 0)


def get_type_arrow(tp: pa.DataType) -> str | None:
    if pa.types.is_decimal(tp):
        return decimal_mssql(tp)
    
    if pa.types.is_timestamp(tp):
        return timestamp_mssql(tp)
    
    if pa.types.is_time(tp):
        return times_mssql(tp)
    
    return None


def get_type_arrow_pyodbc(tp: pa.DataType) -> tuple | None:
    if pa.types.is_decimal(tp):
        return decimal_mssql_pyodbc(tp)
    
    if pa.types.is_timestamp(tp):
        return timestamp_mssql_pyodbc(tp)
    
    if pa.types.is_time(tp):
        return times_mssql_pyodbc(tp)
    
    return None


map_typs = {
    pa.int8(): 'tinyint',
    pa.int16(): 'smallint',
    pa.int32(): 'int',
    pa.int64(): 'bigint',
    pa.bool_(): 'bit',
    pa.float16(): 'real',
    pa.float32(): 'real',
    pa.float64(): 'float',
    pa.date32(): 'date',
    pa.date64(): 'date',
    pa.string(): 'varchar(max)',
    pa.utf8(): 'varchar(max)',
    pa.binary(): 'binary',
}

map_typs_pyodbc = {
    pa.int8(): pyodbc.SQL_TINYINT,
    pa.int16(): pyodbc.SQL_SMALLINT,
    pa.int32(): pyodbc.SQL_INTEGER,
    pa.int64(): pyodbc.SQL_BIGINT,
    pa.bool_(): pyodbc.SQL_BIGINT,
    pa.float16(): pyodbc.SQL_REAL,
    pa.float32(): pyodbc.SQL_REAL,
    pa.float64(): pyodbc.SQL_FLOAT,
    pa.date32(): pyodbc.SQL_TYPE_DATE,
    pa.date64(): pyodbc.SQL_TYPE_DATE,
    pa.string(): (pyodbc.SQL_VARCHAR, 8_000, 0),
    pa.utf8(): (pyodbc.SQL_VARCHAR, 8_000, 0),
    pa.binary(): pyodbc.SQL_BINARY,
}